"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = _default;
exports.recordResetSubfield6OccurrenceNumbers = recordResetSubfield6OccurrenceNumbers;
var _debug = _interopRequireDefault(require("debug"));
var _utils = require("./utils");
var _subfield6Utils = require("./subfield6Utils");
function _interopRequireDefault(e) { return e && e.__esModule ? e : { default: e }; }
// Relocated from melinda-marc-record-merge-reducers (and renamed)

const debug = (0, _debug.default)('@natlibfi/marc-record-validators-melinda:reindexSubfield6OccurrenceNumbers');

// NB! This validator/fixer has two functionalities:
// 1) normal reindexing of occurrence numbers
// 2) disambiguation (when possible) of unambiguous occurrence numbers

function _default() {
  return {
    description: 'Reindex occurrence numbers in $6 subfield so that they start from 01 and end in NN',
    validate,
    fix
  };
  function fix(record) {
    (0, _utils.nvdebug)('Fix SF6 occurrence numbers', debug);
    const res = {
      message: [],
      fix: [],
      valid: true
    };
    //message.fix = []; // eslint-disable-line functional/immutable-data

    // This can not really fail...

    recordDisambiguateSharedSubfield6OccurrenceNumbers(record);
    recordResetSubfield6OccurrenceNumbers(record);

    // message.valid = !(message.message.length >= 1); // eslint-disable-line functional/immutable-data
    return res;
  }
  function validate(record) {
    const res = {
      message: []
    };
    (0, _utils.nvdebug)('Validate SF6 occurrence number multiuses', debug);
    if (recordGetSharedOccurrenceNumbers(record).length) {
      // eslint-disable-line functional/no-conditional-statements
      res.message.push(`Multi-use of occurrence number(s) detected`); // eslint-disable-line functional/immutable-data
    }

    // Check max, and check number of different indexes
    (0, _utils.nvdebug)('Validate SF6 occurrence number (max vs n instances)', debug);
    const max = (0, _subfield6Utils.recordGetMaxSubfield6OccurrenceNumberAsInteger)(record);
    const size = recordGetNumberOfUniqueSubfield6OccurrenceNumbers(record);
    if (max !== size) {
      // eslint-disable-line functional/no-conditional-statements
      res.message.push(`Gaps detected in occurrence numbers: found ${size}, seen max ${max}`); // eslint-disable-line functional/immutable-data
    }
    res.valid = res.message.length < 1; // eslint-disable-line functional/immutable-data
    return res;
  }
}
function getPotentialSharedOccurrenceNumberFields(occurrenceNumber, fields) {
  return fields.filter(f => f.tag !== '880' && f.subfields.some(sf => (0, _subfield6Utils.subfield6GetOccurrenceNumber)(sf) === occurrenceNumber));
}
function subfieldHasSharedOccurrenceNumber(subfield, candFields) {
  const occurrenceNumber = (0, _subfield6Utils.subfield6GetOccurrenceNumber)(subfield);
  if (!occurrenceNumber || occurrenceNumber === '00') {
    return false;
  }
  const relevantFields = getPotentialSharedOccurrenceNumberFields(occurrenceNumber, candFields);
  // record.fields.filter(f => f.tag !== '880' && fieldHasOccurrenceNumber(f, occurrenceNumber));
  return relevantFields.length > 1;
}
function fieldHasSharedOccurrenceNumber(field, candFields) {
  if (!field.subfields || field.tag === '880') {
    // Should not happen
    return false;
  }

  // What if there are multiple $6s in a given field? Should not be, but...
  return field.subfields.some(subfield => subfieldHasSharedOccurrenceNumber(subfield, candFields));
}
function recordGetSharedOccurrenceNumbers(record) {
  const fieldsContainingSubfield6 = record.fields.filter(field => field.tag !== '880' && (0, _utils.fieldHasSubfield)(field, '6'));
  // fieldsContainingSubfield6.some(field => fieldHasSharedOccurrenceNumber(field, fieldsContainingSubfield6)))
  return fieldsContainingSubfield6.filter(field => fieldHasSharedOccurrenceNumber(field, fieldsContainingSubfield6));
}
function recordDisambiguateSharedSubfield6OccurrenceNumbers(record) {
  const sharedOccurrenceNumberFields = recordGetSharedOccurrenceNumbers(record);
  if (sharedOccurrenceNumberFields.length < 2) {
    return;
  }
  (0, _utils.nvdebug)(`Disambiguate occurrence numbers (N=${sharedOccurrenceNumberFields.length}) in...`, debug);
  sharedOccurrenceNumberFields.forEach(field => disambiguateOccurrenceNumber(field));
  function disambiguateable(field) {
    if (field.tag === '880') {
      // Not needed, already filtered...
      return false;
    }
    const occurrenceNumber = (0, _subfield6Utils.fieldGetUnambiguousOccurrenceNumber)(field);
    (0, _utils.nvdebug)(` Trying to disambiguate ${occurrenceNumber} in '${(0, _utils.fieldToString)(field)}`);
    if (occurrenceNumber === undefined) {
      return false;
    }
    const allRelevantFields = getPotentialSharedOccurrenceNumberFields(occurrenceNumber, sharedOccurrenceNumberFields);
    if (allRelevantFields.length < 2) {
      (0, _utils.nvdebug)(` Currently only ${allRelevantFields.length} field(s) use occurrence number ${occurrenceNumber}. No action required.`);
      return false;
    }
    (0, _utils.nvdebug)(` Currently ${allRelevantFields.length} field(s) use occurrence number ${occurrenceNumber}. ACTION REQUIRED!`);
    const relevantFieldsWithCurrFieldTag = allRelevantFields.filter(candField => field.tag === candField.tag);
    if (relevantFieldsWithCurrFieldTag.length !== 1) {
      (0, _utils.nvdebug)(` Number of them using tag ${field.tag} is ${relevantFieldsWithCurrFieldTag.length}. Can not disambiguate!`);
      return false;
    }
    return true;
  }
  function disambiguateOccurrenceNumber(field) {
    if (!disambiguateable(field)) {
      return;
    }
    // Reset field:
    const occurrenceNumber = (0, _subfield6Utils.fieldGetUnambiguousOccurrenceNumber)(field);
    const newOccurrenceNumberAsInt = (0, _subfield6Utils.recordGetMaxSubfield6OccurrenceNumberAsInteger)(record) + 1;
    const newOccurrenceNumber = (0, _subfield6Utils.intToOccurrenceNumberString)(newOccurrenceNumberAsInt);
    const pairedFields = (0, _subfield6Utils.fieldGetOccurrenceNumberPairs)(field, record.fields);
    (0, _utils.nvdebug)(` Reindex '${(0, _utils.fieldToString)(field)}' occurrence number and it's ${pairedFields.length} pair(s) using '${newOccurrenceNumber}'`, debug);
    (0, _subfield6Utils.fieldResetOccurrenceNumber)(field, newOccurrenceNumber, occurrenceNumber);
    pairedFields.forEach(pairedField => (0, _subfield6Utils.fieldResetOccurrenceNumber)(pairedField, newOccurrenceNumber, occurrenceNumber));
  }
}
function recordGetNumberOfUniqueSubfield6OccurrenceNumbers(record) {
  // Calculates the number of used different occurrence numbers
  /* eslint-disable */
  let indexArray = [];
  record.fields.forEach(field => gatherFieldData(field));
  function gatherFieldData(field) {
    if (!field.subfields) {
      return;
    }
    field.subfields.forEach(subfield => gatherSubfieldData(subfield));
  }
  function gatherSubfieldData(subfield) {
    if (!(0, _subfield6Utils.isValidSubfield6)(subfield)) {
      return;
    }
    const i = (0, _subfield6Utils.subfield6GetOccurrenceNumberAsInteger)(subfield);
    if (i === 0) {
      return;
    }
    indexArray[i] = 1;
  }
  let n = 0;
  indexArray.forEach(elem => n += elem);
  /* eslint-enable */
  return n;
}
function recordResetSubfield6OccurrenceNumbers(record) {
  // Remove gaps
  /* eslint-disable */
  let currentInt = 1;
  let oldtoNewCache = {};
  record.fields.forEach(field => fieldResetSubfield6(field));
  function fieldResetSubfield6(field) {
    (0, _utils.nvdebug)(`fieldResetSubfield6(${(0, _utils.fieldToString)(field)}), CURR:${currentInt}`, debug);
    if (!field.subfields) {
      return;
    }
    field.subfields.forEach(subfield => subfieldReset6(subfield));
  }
  function subfieldReset6(subfield) {
    if (!(0, _subfield6Utils.isValidSubfield6)(subfield)) {
      return;
    }
    const currIndex = (0, _subfield6Utils.subfield6GetOccurrenceNumber)(subfield);
    if (currIndex === undefined || currIndex === '00') {
      return;
    }
    const newIndex = mapCurrIndexToNewIndex(currIndex);
    //nvdebug(`subfieldReset6(${subfieldToString(subfield)}): ${newIndex}`, debug);
    (0, _subfield6Utils.subfield6ResetOccurrenceNumber)(subfield, newIndex);
  }
  function mapCurrIndexToNewIndex(currIndex) {
    if (currIndex in oldtoNewCache) {
      return oldtoNewCache[currIndex];
    }
    const newIndex = (0, _subfield6Utils.intToOccurrenceNumberString)(currentInt);
    oldtoNewCache[currIndex] = newIndex;
    currentInt++;
    return newIndex;
  }

  /* eslint-enable */
}
//# sourceMappingURL=reindexSubfield6OccurenceNumbers.js.map